/*
    SequencerALSA.h
  
    Copyright (C) 1998, 1999 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef SequencerALSA_H
#define SequencerALSA_H

#include <sys/asoundlib.h>

#include "util/mytype.h"
#include "Sequencer/SequencerImple.h"
#include "AlsaClient.h"

class SequencerALSA:public SequencerImple
{
protected:
  //100HzưäˤǤʤФʤʤϤʤ
  static const int CONTROL_RATE = 100;
  //static const int CONTROL_RATE = 1024;//Ϥä̵?
  
  virtual bool checkPortChannel (MIDI::PORT port, uchar channel);


  ///////////////////////////////////////////////////////////ALSA stuff
  AlsaClient* alsaClient;
  
  static const int DEFAULT_DEST_QUEUE = 0;//first queue
  static const int DEFAULT_DEST_CLIENT_A = 64;//first kernel sequencer client
  static const int DEFAULT_DEST_PORT_A = 0;//first port 
  static const int DEFAULT_DEST_CLIENT_B = 64;//first kernel sequencer client
  static const int DEFAULT_DEST_PORT_B = 1;//second port

  static const char DEST_QUEUE[] = "WIDI_SEQUENCER_ALSA_DEST_QUEUE";
  static const char DEST_CLIENT_A[] = "WIDI_SEQUENCER_ALSA_DEST_CLIENT_A";
  static const char DEST_PORT_A[] = "WIDI_SEQUENCER_ALSA_DEST_PORT_A";
  static const char DEST_CLIENT_B[] = "WIDI_SEQUENCER_ALSA_DEST_CLIENT_B";
  static const char DEST_PORT_B[] = "WIDI_SEQUENCER_ALSA_DEST_PORT_B";

  //ϴĶѿǷꤹ롣
  int srcPort;
  const int destQueue;
  const int destClientA;
  const int destPortA;
  bool isPortASubscribing;
  bool isPortAUsable;
  const int destClientB;
  const int destPortB;
  bool isPortBSubscribing;
  bool isPortBUsable;
  
  
  class DestQueueClient
  {
  private:
    int queue;
    int destClient;
    int destPort;
    int destChannel;
    
  public:
    DestQueueClient ():
      queue (-1), destClient (-1), destPort (-1), destChannel (-1)
      {};
    DestQueueClient (int queue__, int destClient__,
		     int destPort__, int destChannel__):
      queue (queue__), destClient (destClient__),
      destPort (destPort__), destChannel (destChannel__)
      {};
    
    void set (int queue__, int destClient__, int destPort__, int destChannel__)
      {
	queue = queue__;
	destClient = destClient__;
	destPort = destPort__;
	destChannel = destChannel__;
      };
    
    int getQueue () const {assert (queue != -1); return queue;};
    int getDestClient () const {assert (destClient != -1); return destClient;};
    int getDestPort () const {assert (destPort != -1); return destPort;};
    int getDestChannel () const
      {assert (destChannel != -1); return destChannel;};
  };

  //(port, channel)->(alsa queue, alsa client, alsa port, alsa channel)
  //ؤѴơ֥
  DestQueueClient destQueueClient[MIDI::MAX_PORT][MIDI::MAX_CHANNEL];
  //kernel clientäsubscribe롣
  //user clientäsubscribeʤsubscribe롣
  //ʤФǤɤ
  //̤isSubscribing, isUsable֤
  void initializePort (int srcPort__,
		       int destQueue__, int destClient, int destPort,
		       bool* isSubscribing, bool* isUsable);
  //subscribeƤʤunsubscribe롣
  void releasePort (int srcPort__,
		    int destQueue__, int destClient, int destPort,
		    bool isSubscribing);

  
  //machine tickssnd_seq_real_time_t׻롣
  void calcTicks2SndSeqRealTime (ulong machineTicks__,
				 snd_seq_real_time_t* sndSeqRealTime__)
    {
      sndSeqRealTime__->tv_sec = machineTicks__/getCtrlRate ();
      //1000000000.0nano second(10^9)
#if 0
      sndSeqRealTime__->tv_nsec =
	static_cast<int>(1000000000.0*(machineTicks__-sndSeqRealTime__->tv_sec*getCtrlRate ())/getCtrlRate ());
#else
      sndSeqRealTime__->tv_nsec =
	static_cast<int>((machineTicks__-sndSeqRealTime__->tv_sec*getCtrlRate ())*(1000000000.0/getCtrlRate ()));
#endif
    };
  
  //ߤalsa real time(machine ticks׻롣)
  //Ūˤevent˷׻ľɤ,̵̤ʷ׻򸺤餹ᡣ
  snd_seq_real_time_t currentRealTime;
  
public:
  SequencerALSA (State* globalState, Option* option__);
  virtual ~SequencerALSA ();

  virtual void reset ();
  
  virtual int getCtrlRate () throw (Error) {return CONTROL_RATE;};
  virtual void sync ();

protected:
  virtual void doDumpBuf ();

  virtual bool doPortChange (MIDI::PORT port__) {return true;};
  
  //ºݤ˥٥Ȥroutine
  virtual void doNoteOff (int port, uchar channel,
			  uchar note, uchar velocity);
  virtual void doNoteOn (int port, uchar channel,
			 uchar note, uchar velocity);
  virtual void doPolyphonicKeyPressure (int port, uchar channel,
					uchar note,
					uchar keyPressure);
  virtual void doControlChange (int port, uchar channel,
				uchar control, uchar value);
  virtual void doProgramChange (int port, uchar channel, uchar  program);
  virtual void doChannelPressure (int port, uchar channel,
				  uchar channelPressure);
  virtual void doPitchBendChange (int port, uchar channel,
				  uchar lsb, uchar msb);
  virtual void doExclusive (const uchar* udata, uint32 size,
			    int port =-1, int channel=-1);

  //
  //timerϢ
  //
  virtual void doWaitTimer (ulong ticks) {};
  virtual void doDeltaTimer (ulong ticks) {};
  virtual void doStartTimer ();
  virtual void doStopTimer ();
  
protected:
  virtual void traceWaitTimer (ulong waitForMachineTicks)
    {
      Sequencer::traceWaitTimer (waitForMachineTicks);
      calcTicks2SndSeqRealTime (machineTicks, &currentRealTime);
    };
  virtual void traceDeltaTimer (ulong waitForMachineTicks)
    {
      Sequencer::traceDeltaTimer (waitForMachineTicks);
      calcTicks2SndSeqRealTime (machineTicks, &currentRealTime);
    }
  virtual void traceStartTimer ()
    {
      Sequencer::traceStartTimer ();
      calcTicks2SndSeqRealTime (machineTicks, &currentRealTime);
    };
  
public:
  //exceptional classes
  class AError:public SequencerError {};
  class CanNotOpenAlsaDevice:public CanNotOpenSequencerDevice {};
  class CanNotCreatePort:public SequencerError {};
  class CanNotFindUsablePort:public SequencerError {};
};
#endif // SequencerALSA_H
