/*
    DotData.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cassert>

#include "DotData.h"

const int DotData::MAX_LINES;
const int DotData::MAX_COLS;
const int DotData::ROLAND_DOT_DATA_SIZE;

DotData::DotData ()
{
  reset ();
}

DotData::~DotData ()
{
  //do nothing
}

void DotData::reset ()
{
  for (int y = 0; y < MAX_LINES; y++)
    {
      for (int x = 0; x < MAX_COLS; x++)
	__isOn[y][x] = false;
    }
}


inline void DotData::setRolandLCD (const uchar rolandDotData[ROLAND_DOT_DATA_SIZE],
				   bool lcd[MAX_LINES][MAX_COLS],
				   const int dataOffset,
				   const int columnOffset)
{
  for (int y = 0; y < 16; y++)
    {
      for (int x = 0; x < 5; x++)
        {
          if (((rolandDotData[y+dataOffset])&(0x10 >> x)) != 0)
	    lcd[y][x+columnOffset] = true;
        }
    }
}


void DotData::setRolandDotData (const uchar rolandDotData[ROLAND_DOT_DATA_SIZE])
{
  reset ();
  
  ///lcd$B$N9=@.(B
  //dotData 00-15
  setRolandLCD (rolandDotData, __isOn, 0, 0);
  //dotData 16-31
  setRolandLCD (rolandDotData, __isOn, 16, 5);
  //dotData 16-31
  setRolandLCD (rolandDotData, __isOn, 32, 10);
  //dotData 48-63
  for (int y = 0; y < 16; y++)
    {
      if ((rolandDotData[y+48] & 0x10))
	__isOn[y][15]=true;
    }
}
