/*
    SequencerEmulation.cc
  
    Copyright (C) 1999 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#include "SequencerEmulation.h"

SequencerEmulation::SequencerEmulation (DlHandler** dlHandler__,
					const string& name,
					State* globalState,
					Option* midiOption):
  SequencerImple (globalState, midiOption)
{
  //$B<B:]$K(Bmidi event$B$rAw?.$9$k(Bsequencer$B$r(Bopen$B$9$k!#(B
  try
    {
#ifdef USE_DUMMBY_STATE
      sequencer = newSequencer (dlHandler__, name, &dummyState, midiOption);
#elif defined (USE_GLOBAL_STATE)
      sequencer = newSequencer (dlHandler__, name, &cacheState, midiOption);
#endif
    }
  catch (Sequencer::SequencerError&)
    {
      cerr << "SeqeuncerConversion::SequencerEmulation: "
	   << "can not create seqeucner" << endl;
      throw;//$B:F%9%m!<(B
    }

  
}

SequencerEmulation::~SequencerEmulation ()
{
  delete sequencer;
}



void SequencerEmulation::doNoteOff (int port, uchar channel,
				    uchar note, uchar velocity)
{
  sequencer->noteOff (port, channel, note, velocity);
}

void SequencerEmulation::doNoteOn (int port, uchar channel,
				   uchar note, uchar velocity)
{
  sequencer->noteOn (port, channel, note, velocity);
}

void SequencerEmulation::doPolyphonicKeyPressure (int port, uchar channel,
						  uchar note,
						  uchar keyPressure)
{
  sequencer->polyphonicKeyPressure (port, channel, note, keyPressure);
}

void SequencerEmulation::doControlChange (int port, uchar channel,
					  uchar control, uchar value)
{
  //XXX
  sequencer->controlChange (port, channel, control, value);
}

void SequencerEmulation::doProgramChange (int port, uchar channel,
					  uchar program)
{
  //XXX
  sequencer->programChange (port, channel, program);
}

void SequencerEmulation::doChannelPressure (int port, uchar channel,
					    uchar channelPressure)
{
  sequencer->channelPressure (port, channel, channelPressure);
}

void SequencerEmulation::doPitchBendChange (int port, uchar channel,
					    uchar lsb, uchar msb)
{
  sequencer->pitchBendChange (port, channel, lsb, msb);
}

void SequencerEmulation::doExclusive (const uchar* udata, uint32 size,
				      int port =-1, int channel=-1)
{
  //XXX
  sequencer->exclusive (udata, size, port, channel);
}

