/*
    SequencerEmulation.h
  
    Copyright (C) 1999 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */

#ifndef SequencerEmulation_H
#define SequencerEmulation_H

#include "Sequencer/SequencerImple.h"
#include "EmulationConfig.h"

//midi emulation$B$r<B9T$9$k%/%i%9(B
class SequencerEmulation:public SequencerImple
{
private:

  //$B<B:]$N1iAU$K;HMQ$9$k(Bseqeuncer device
#define USE_DUMMY_STATE
  //#define USE_GLOBAL_STATE

#ifdef USE_DUMMY_STATE
  State dummyState;//seqeuncer$B$KEO$9$?$a$N%@%_!<$NJ*!#(B???
#endif
  Sequencer* sequencer;
  
  //midi event conversion$B$N@_Dj!#(B
  EmulationConfig emulationConfig;

public:
  SequencerEmulation (DlHandler** dlHandler__, const string& name,
		      State* globalState, Option* midiOption);
  ~SequencerEmulation ();

#if defined (USE_DUMMY_STATE)
  //$B<h$j4:$($:(B
  //no need to override updateStatus ()
  //$B$H$7$F$*$/$,(Bplaying status indicator$B$KBP$7$F$I$N>u67$rI=<($7$?$iNI$$$N(B
  //$B$@$m$&(B?
  //$B0?$O%*%W%7%g%s$GA*$Y$k$h$&$K$9$Y$-(B?
#elif defined (USE_GLOBAL_STATE)
  //globalState$B$r(Bupdate$B$9$k!#$3$l$O%m%C%/$7$J$$$N$G8F$S=P$7B&$G(B
  //$B%m%C%/$7$F$*$/I,MW$,$"$k!#(B
  virtual void updateStatus ()
    {
      //sequencer$B$N>uBV$r(BglobalState$B$K(Bcopy$B$9$k!#(B
      sequencer->updateStatus ();
    };
#endif

  
  //device$B$r(Breset$B$9$k!#(B
  virtual void reset ()
    {
      sequencer->reset ();
      traceReset ();
    };
  virtual void allNoteOffSoundOff ()
    {
      sequencer->allNoteOffSoundOff ();
      traceAllNoteOffSoundOff ();
    };
  //
  //timer$B4XO"!#(B
  //
  virtual int getCtrlRate () throw (Error)
    {
      return sequencer->getCtrlRate ();
    };

  //sync
  virtual void sync ()
    {
      sequencer->sync ();
    };
  
protected:
  virtual void doDumpBuf ()
    {
      sequencer->dumpBuf ();
    };

  virtual bool doPortChange (MIDI::PORT port__)
    {
      sequencer->portChange (port__);
      return true;//XXX???
    };
  
  //$B<B:]$K%$%Y%s%H$r=hM}$9$k(Broutine
  virtual void doNoteOff (int port, uchar channel, uchar note, uchar velocity);
  virtual void doNoteOn (int port, uchar channel, uchar note, uchar velocity);
  virtual void doPolyphonicKeyPressure (int port, uchar channel,
					uchar note,
					uchar keyPressure);
  virtual void doControlChange (int port, uchar channel,
				uchar control, uchar value);
  virtual void doProgramChange (int port, uchar channel, uchar  program);
  virtual void doChannelPressure (int port, uchar channel,
				  uchar channelPressure);
  virtual void doPitchBendChange (int port, uchar channel,
				  uchar lsb, uchar msb);
  virtual void doExclusive (const uchar* udata, uint32 size,
			    int port =-1, int channel=-1);

  //
  //timer$B4XO"!#(B
  //
  virtual void doWaitTimer (ulong ticks)
    {
      sequencer->waitTimer (ticks);
    };
  virtual void doDeltaTimer (ulong ticks)
    {
      sequencer->deltaTimer (ticks);
    };
  //$B$3$l$O%?%$%^!<$r(B0$B$K$9$k!#(B
  virtual void doStartTimer ()
    {
      sequencer->startTimer ();
    };
  virtual void doStopTimer ()
    {
      sequencer->stopTimer ();
    };

protected:
#ifdef USE_DUMMY_STATE
  //no need to override to traceReset ()
#else // USE_GLOBAL_STATE 
  virtual void traceReset ()
    {
      resetPartData ();
      cacheState.sequencerStatus.reset ();
      traceStartTimer ();
      //!!! CAUSTION !!!
      //globalState$B$H$7$FF1$8$b$N$r(Bseqeuncer$B$KEO$7$?$i(Bseqeuncer->state$B$G(B
      //state->lock ();
      //updateStatus ();
      //state->broadcast ();
      //state->unlock ();
      //$B$r<B9T$9$k$N$G!"$3$3$G<B9T$9$kI,MW$O$J$$!#(B
      //$B:G8e$N(BstartTimer ();
      //$B$b(Bseqeuncer->startTimer ()$B$r8F$S=P$9I,MW$,$J$$$N$G(B,
      //traceStartTimer ()$B$K$J$C$F$$$k$3$H$KCm0U!#(B
    };
#endif
};

#endif // SequencerEmulation_H
