/*
    Part.cc
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */


#include <cstdio>
#include <cassert>

//#include "instrument.h"
#include "Part.h"
//#include "InstrumentName.h"

using namespace std;

InstrumentName* Part::instrumentName = NULL;
pthread_once_t Part::isInstrumentInited = PTHREAD_ONCE_INIT;

const bool Part::ON;
const bool Part::OFF;
const uchar Part::MAP_OFF;
const uchar Part::MAP_1;
const uchar Part::MAP_2;
const int32 Part::MAX_PITCH_BEND;
const int32 Part::CENTER_PITCH_BEND;
const uint32 Part::MAX_MODULATION;
const uint32 Part::MAX_CHANNEL_VOLUME;
const uint32 Part::MAX_PAN;
const uint32 Part::CENTER_PAN;
const uint32 Part::MAX_EXPRESSION;
const uchar Part::MAX_REVERB_SEND_LEVEL;
const int Part::PARAMETER_NULL;
const int Part::PARAMETER_RPN;
const int Part::PARAMETER_NRPN;

void Part::initInstrumentName ()
{
  assert (instrumentName == NULL);
  instrumentName = new InstrumentName ();
}


Part::Part ()
{
  pthread_once (&isInstrumentInited, &Part::initInstrumentName);
  reset (MIDI::PORT_A, 0);
}

Part::~Part ()
{
  //delete instrumentName;
}


void Part::reset (uchar defaultPort, uchar receivedChannel)
{
  //assert (0<=receivedChannel);
  assert (receivedChannel < MIDI::MAX_CHANNEL);
#if 0
  assert (defaultPort == SequencerStatus::PORT_A ||
	  defaultPort == SequencerStatus::PORT_B);
#else
  //assert (0 <= defaultPort);
  assert (defaultPort < MIDI::MAX_PORT);
#endif
  RxChannel = receivedChannel;
  if (receivedChannel == 10-1)
    useForRhythmPart = MAP_1;
  else
    useForRhythmPart = MAP_OFF;
  port = defaultPort;

  reset ();
}

void Part::reset ()
{
  //XXX sc88_pro$B$G$OBLL\!#@_Dj$K$h$jJQ$($k$3$H!#(B
  if (useForRhythmPart == MAP_OFF)
    //instrument = inst_sc88_pro[0][0];
    instrument = instrumentName->getDefaultInstrumentName ();
  else
    //instrument = drum_sc88_pro[0];
    instrument = instrumentName->getDefaultDrumName ();
  
  __isMuted = false;
  //XXX:default$B$NCM$OE,Ev!#8e$G!"$A$c$s$H$7$?CM$KD>$9$3$H!#(B
  for (int i = 0; i < MIDI::MAX_NOTES; i++)
    {
      lastNote[i] = OFF;
      note[i] = OFF;
      velocity[i] = 0;
      polyphonicKeyPressure[i] = 0;
      noteOnMachineTicks[i] = 0;
    }
  
  maxVelocity = 0;
  lastMaxVelocity = 0;
  lastMachineTicks = 0;
  notesOnNumber = 0;
  channelPressure = 0;
  program = 0;
  bankMSB = 0;
  bankLSB = 0;
  
  pitchBend = 0;
  modulationWheel = 0;
  channelVolume = MAX_CHANNEL_VOLUME;
  pan = CENTER_PAN;
  expression = MAX_EXPRESSION;  
  damperPedal = 0;
  portamentSwitch = 0;
  sostenute = 0;
  softPedal = 0;

  reverbSendLevel = 0;
  chorusSendLevel = 0;
  delaySendLevel = 0;

  parameterKind = PARAMETER_NULL;
  rpnMSB = 0;
  rpnLSB = 0;
  nrpnMSB = 0;
  nrpnLSB = 0;
  vibrateRate = 0;
  vibrateDepth = 0;
  vibrateDelay = 0;
  tvfCutOff = 0;
  tvfResonance = 0;
  envelopeAttackTime = 0;
  envelopeDecayTime = 0;
  envelopeReleaseTime = 0;

#if 0
  toneNumber = ;
  RxPitchBend = ;
  RxChPressure = ;
  RxProgramChange = ;
  RxCongrolChange = ;
  RxPolyPressure = ;
  RxNoteMessage = ;
  RxRPN = ;
  RxNRPN = ;
  RxModulation = ;
  RxVolume = ;
  RxPanpot = ;
  RxExpression = ;
  RxHold1 = ;
  RxPortamento = ;
  RxSostenute = ;
  RxSoft = ;
  MonoPolyMode = ;
  assignMode = ;

  RxBankSelect = ;
  RxBandSelectLSB = ;
#endif
}


uchar Part::getNoteOnVelocity (const SongStatus& nowSongStatus) const
{
  uchar retVelocity = 0;
  for (int note = 0; note < MIDI::MAX_NOTES; note++)
    {
      if (retVelocity < velocity[note])
        retVelocity = velocity[note];
    }
#define DELAY (nowSongStatus.getRate ()*120/100)//1.2 second
#define DECAY_RATE 40/100 //40%$BA4$/:,5r$J$7!#;n9T:x8m$G7h$a$k!#(B
  if (nowSongStatus.getMachineTicks ()-lastMachineTicks >= DELAY)
    return retVelocity*DECAY_RATE;
  else
    return retVelocity;
#undef DECAY_RATE
#undef DELAY
}

uchar Part::getMaxVelocity (const SongStatus& nowSongStatus) const
{
#define DELAY (nowSongStatus.getRate ()*15/100)//0.15 second
  if (nowSongStatus.getMachineTicks ()-lastMachineTicks < DELAY)
    return lastMaxVelocity;
  else
    return 0;
#undef DELAY
}


uchar Part::getNoteOnVelocity2 (const SongStatus& nowSongStatus) const
{
  uchar retVelocity = 0;
  for (int note = 0; note < MIDI::MAX_NOTES; note++)
    {
      if (retVelocity < velocity[note])
        retVelocity = velocity[note];
    }
#define DELAY (nowSongStatus.getRate ()*40/100)//0.4 second
  if (nowSongStatus.getMachineTicks ()-lastMachineTicks >= DELAY)
    return 0;
  else
    return retVelocity;
#undef DELAY
}

uchar Part::getMaxVelocity2 (const SongStatus& nowSongStatus) const
{
#define DELAY (nowSongStatus.getRate ()*15/100)//0.15 second
  if (nowSongStatus.getMachineTicks ()-lastMachineTicks < DELAY)
    return lastMaxVelocity;
  else
    return 0;
#undef DELAY
}


void Part::traceNoteOff (uchar note__)
{
  assert (note__ < 128);

  lastNote[note__] = note[note__];
  note[note__] = Part::OFF;
  velocity[note__] = 0;
  notesOnNumber--;
  if (notesOnNumber < 0)
    notesOnNumber = 0;
}

void Part::traceNoteOn (uchar note__, uchar velocity__,
			    double machineTicks)
{
  assert (note__ < 128);
  assert (velocity__ < 128);
  
  lastNote[note__] = note[note__];
  velocity[note__] = velocity__;
  noteOnMachineTicks[note__] = machineTicks;
  
  //drum part$BEy$N0Y$K:GBg$NJ*$r3P$($F$*$/!#(B
  //$B$9$0$K(Bnote off$B$9$k>l9g$,$"$k$N$G$=$NBP:v!#(B
  if (maxVelocity < velocity__)
    maxVelocity = velocity__;
  if (velocity__ > 0)
    {
      note[note__] = Part::ON;
      notesOnNumber++;
      lastMachineTicks = static_cast<uint32> (machineTicks);
    }
  else
    {
      traceNoteOff (note__);
    }
}

void Part::tracePolyphonicKeyPressure (uchar note__, uchar keyPressure)
{
  assert (note__ < 128);
  assert (keyPressure < 128);
  
  polyphonicKeyPressure[note__] = keyPressure;
}

namespace
{
  inline void setValue (bool isMSB, uint32& toBeSet, uchar value)
    {
      assert (value < 128);
  
      if (isMSB)
	{
	  toBeSet=(value << 7);
	}
      else
	{
	  toBeSet &= 0xFF70;//0-7bit$B$r(Bclear
	  toBeSet |= value;
	}
    }
  
  inline void setValue (uchar& toBeSet, uchar value)
    {
      assert (value < 128);
      toBeSet = value;
    }
}

void Part::traceControlChange (uchar control, uchar value)
{
  assert (control < 128);
  assert (value < 128);
  
  /////////////////////////////////////////bank select
  if (control == 0)
    {
      bankMSB = value;
      return;
    }
  if (control == 32)
    {
      bankLSB = value;
      return;
    }
  
  
  ///////////////////////////////////////////control<63
  if (control < 63)
    {
      const bool isMSB = (control<32);
      if (!isMSB) control -= 32;
      switch (control)
	{
	case 1:
	  //modulation wheel
	  setValue (isMSB, modulationWheel, value);
	  break;
	case 2:
	  //breath controller
	  break;
	case 4:
	  //foot controller
	  break;
	case 5:
	  //portament time
	  break;
	case 6:
	  //data entry
	  traceDataEntry (isMSB, value);
	  break;
	case 7:
	  //channel volume
	  setValue (isMSB, channelVolume, value);
	  break;
	case 8:
	  //balance
	  break;
	case 10:
	  //pan
	  setValue (isMSB, pan, value);
	  break;
	case 11:
	  //expresion
	  setValue (isMSB, expression, value);
	  break;
	case 12:
	  //effect control 1
	  break;
	case 13:
	  //effect control 2
	  break;
	case 14:
	  break;
	case 15:
	  break;
	case 16:
	  //general purpose controller 1
	  break;
	case 17:
	  //general purpose controller 2
	  break;
	case 18:
	  //general purpose controller 3
	  break;
	case 19:
	  //general purpose controller 4
	  break;
	case 20:
	    break;
	case 21:
	    break;
	case 22:
	    break;
	case 23:
	    break;
	case 24:
	    break;
	case 25:
	    break;
	case 26:
	    break;
	case 27:
	    break;
	case 28:
	    break;
	case 29:
	    break;
	case 30:
	    break;
	case 31:
	    break;
	case 32:
	    break;
	}
      return;
    }

  ////////////////////////////64 <= control <= 95
  if (control <= 95)
    {
      switch (control)
	{
	case 64:
	  //damper pedal (hold 1)
	  setValue (damperPedal, value);
	  break;
	case 65:
	  //portament switch
	  setValue (portamentSwitch, value);
	  break;
	case 66:
	  //sostenuto
	  setValue (sostenute, value);
	  break;
	case 67:
	  //soft pedal
	  setValue (softPedal, value);
	  break;
	case 68:
	  //regard foot switch
	  break;
	case 69:
	  //hold 2
	  break;
	case 70:
	  //sound controller 1
	  //sound variation
	  break;
	case 71:
	  //sound controller 2
	  //timbre/harmonic intensity
	  break;
	case 72:
	  //sound controller 3
	  //release time
	  break;
	case 73:
	  //sound controller 4
	  //attack time
	  break;
	case 74:
	  //sound controller 5
	  //brightness
	  break;
	case 75:
	  //sound controller 6
	  //decay time
	  break;
	case 76:
	  //sound controller 7
	  //vibrato rate
	  break;
	case 77:
	  //sound controller 8
	  //vibrato depth
	  break;
	case 78:
	  //sound controller 9
	  //vibrato delay
	  break;
	case 79:
	  //sound controller 10
	  break;
	case 80:
	  //general purpose controller 5
	  break;
	case 81:
	  //general purpose controller 6
	  break;
	case 82:
	  //general purpose controller 7
	  break;
	case 83:
	  //general purpose controller 8
	  break;
	case 84:
	  //portament controll
	  break;
	case 85:
	  break;
	case 86:
	  break;
	case 87:
	  break;
	case 88:
	  break;
	case 89:
	  break;
	case 90:
	  break;
	case 91:
	  //effect1 depth
	  //reverb send level
	  setValue (reverbSendLevel, value);
	  break;
	case 92:
	  //effect2 depth
	  //tremolo depth
	  break;
	case 93:
	  //effect3 depth
	  //chorus send level
	  setValue (chorusSendLevel, value);
	  break;
	case 94:
	  //effect4 depth
	  //celeste(detune) depth
	  setValue (delaySendLevel, value);
	  break;
	case 95:
	  //effect5 depth
	  //phaser depth
	  break;
	}
      return;
    }

  /////////////////////////////////////////RPN,NRPN
  if (control <= 101)
    {
      switch (control)
	{
	case 96:
	  //increment
	  break;
	case 97:
	  //decrement
	  break;
	case 98:
	  //NRPN LSB
	  setValue (nrpnLSB, value);
	  parameterKind = Part::PARAMETER_NRPN;
	  break;
	case 99:
	  //NRPN MSB
	  setValue (nrpnMSB, value);
	  parameterKind = Part::PARAMETER_NRPN;
	  break;
	case 100:
	  //RPN LSB
	  setValue (rpnLSB, value);
	  parameterKind = Part::PARAMETER_RPN;
	  break;
	case 101:
	  //RPN MSB
	  setValue (rpnMSB, value);
	  parameterKind = Part::PARAMETER_RPN;
	  break;	  
	}
      return;
    }

  ////////////////////////////////////////////////////reserved
  if (control<119)
    {
      return;
    }
  
  /////////////////////////////////////////////////////MODE
  switch (control)
    {
    case 120:
      //all sound off
      traceAllSoundOff ();
      break;
    case 121:
      //rset all controllers
      traceResetAllController ();
      break;      
    case 122:
      //local control
      break;      
    case 123:
      //all note off
      break;
    case 124:
      //omni off
      break;
    case 125:
      //omni on
      break;
    case 126:
      //mono mode on(poly mode off)
      break;
    case 127:
      //poly mode on(mono mode off)
      break;
    }
  return; 
}

namespace
{
  inline void setRelativeValue (char& toBeSet, uchar value)
    {
      int result = toBeSet+value-0x40;
      //-64 <= value <= 63
      //0x40$B$,(B0
      if (result>63) result = 63;
      if (result<-64)result = -64;
      toBeSet = result;
    }
}

void Part::traceDataEntry (bool isMSB, uchar value)
{
  //RPN,NRPN$B$G8+$F$$$k$N$O(BNRPN$B$N(B
  //0x01 0x08,0x01 0x09,0x01 0x0A
  //0x01 0x20,0x01 0x21,
  //0x01 0x63,0x01 0x64,0x01 0x65
  //$B$N$_$J$N$G$=$l0J30$O2?$b$7$J$$$G(Breturn$B!#(B
  if (parameterKind == Part::PARAMETER_NULL||
      parameterKind == Part::PARAMETER_RPN)
    return;
  if (!isMSB)
    return;
  if (nrpnMSB != 0x01)
    return;
  switch (nrpnLSB)
    {
    case 0x08:
      setRelativeValue (vibrateRate, value);
      break;
    case 0x09:
      setRelativeValue (vibrateDepth, value);
      break;
    case 0x0A:
      setRelativeValue (vibrateDelay, value);
      break;
    case 0x20:
      setRelativeValue (tvfCutOff, value);
      break;
    case 0x21:
      setRelativeValue (tvfResonance, value);
      break;
    case 0x63:
      setRelativeValue (envelopeAttackTime, value);
      break;
    case 0x64:
      setRelativeValue (envelopeDecayTime, value);
      break;
    case 0x65:
      setRelativeValue (envelopeReleaseTime, value);
      break;
    default:
#ifdef DEBUG
      //cerr << "DEBUG:unkown NRPN!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl;
#endif
      break;
    }
}

void Part::traceAllSoundOff ()
{
  for (int i = 0; i < MIDI::MAX_NOTES; i++)
    {
      lastNote[i] = Part::OFF;
      note[i] = Part::OFF;
      velocity[i] = 0;
    }
  maxVelocity = 0;
  lastMaxVelocity = 0;
  notesOnNumber = 0;
}

void Part::traceResetAllController ()
{
  pitchBend = 0;
  for (int i = 0; i < MIDI::MAX_NOTES; i++)
    polyphonicKeyPressure[i] = 0;
  channelPressure = 0;
  modulationWheel = 0;
  expression = Part::MAX_EXPRESSION;
  damperPedal = 0;
  portamentSwitch = 0;
  sostenute = 0;
  softPedal = 0;
  parameterKind = Part::PARAMETER_NULL;
  rpnMSB = 0x7F;
  rpnLSB = 0x7F;
}



namespace
{
  inline void setInstrument (const char*& toBeSet, const char* instrumentName)
    {
      if (instrumentName != NULL)
	toBeSet = instrumentName;
      else
	toBeSet = "------";//no instrument is assinged.
    }
}

void Part::traceProgramChange (uchar program__,
			       SequencerStatus::DeviceMode deviceMode)
{
  assert (program__ < 128);
  
  //$BK\Ev$O(BXG$B$H$+$K$bBP1~$9$Y$-$@$,!#(B
  program = program__;
#if 0
  assert (bankLSB == 0 ||
	  bankLSB == 1 ||
	  bankLSB == 2 ||
	  bankLSB == 3);
  assert (useForRhythmPart == Part::MAP_OFF ||
	  useForRhythmPart == Part::MAP_1   ||
	  useForRhythmPart == Part::MAP_2);
#endif

  //$B0J2<(Binstrument$B$rBeF~!#(B
  setInstrument (instrument,
		instrumentName->getInstrumentName (program,
						   bankLSB, bankMSB,
						   deviceMode,
						   useForRhythmPart));
#if 0
  if ((deviceMode&SequencerStatus::DEVICE_XG_ON) != 0)
    {
      //???!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      if (bankMSB == 0)
	setInstrument (instrument, inst_xg[program][bankLSB]);
      else if (bankMSB == 64)
	setInstrument (instrument, sfx_xg[program]);
      else if (bankMSB == 126)
	setInstrument (instrument, drum_sfx_xg[program]);
      else if (bankMSB == 127)
	setInstrument (instrument, drum_xg[program]);
    }
  else
    {
      switch (bankLSB)
	{
	case 0:
	  //$B5!3#$N(Bbutton$B$G@_Dj$5$l$F$$$k(Bmap$B$K$9$k!#2r$i$J$$$N$G<h$j4:$($:(B88pro map$B$K$7$F$*$/!#(B
	  //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	  if (useForRhythmPart == Part::MAP_OFF)
	    setInstrument (instrument, 
			   inst_sc88_pro[program][bankMSB]);
	  else
	    setInstrument (instrument, drum_sc88_pro[program]);
	  break;
	case 1://sc-55 map
	  if (useForRhythmPart == Part::MAP_OFF)
	    setInstrument (instrument, 
			   inst_sc88_55[program][bankMSB]);
	  else
	    setInstrument (instrument, drum_sc88_55[program]);
	  break;
	case 2://sc-88 map
	  if (useForRhythmPart == Part::MAP_OFF)
	    setInstrument (instrument,
			   inst_sc88_88[program][bankMSB]);
	  else
	    setInstrument (instrument, drum_sc88_88[program]);
	  break;
	case 3://sc-88pro map
	  if (useForRhythmPart == Part::MAP_OFF)
	    setInstrument (instrument,
			   inst_sc88_pro[program][bankMSB]);
	  else
	    setInstrument (instrument, drum_sc88_pro[program]);
	  break;
	}
    }
#endif
}

void Part::traceChannelPressure (uchar channelPressure__)
{
  assert (channelPressure < 128);
  
  channelPressure = channelPressure__;
}

void Part::tracePitchBendChange (uchar LSB, uchar MSB)
{
  assert (LSB < 128);
  assert (MSB < 128);
  
  pitchBend = (MSB << 7)+LSB-8191;
  //$B<h$kHO0O$O(B
  //0x00 0x00H --- 0x40 0x00 --- 0x7F 0x7F (-8192 --- 0 --- 8191)
}
