/*
    Part.h
  
    Copyright (C) 1998 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    */



#ifndef Part_H
#define Part_H
#include <pthread.h>

#include "Song/SongStatus.h"
#include "Sequencer/SequencerStatus.h"
#include "MIDI.h"
#include "InstrumentName.h"

class Part
{
public:
  //note
  static const bool ON = true;
  static const bool OFF = false;

  //map
  static const uchar MAP_OFF = 0;
  static const uchar MAP_1   = 1;
  static const uchar MAP_2   = 2;

  //pitch bend
  static const int32 MAX_PITCH_BEND = 8192;
  static const int32 CENTER_PITCH_BEND = 0;

  ///0xBn
  static const uint32 MAX_MODULATION = 16383;
  static const uint32 MAX_CHANNEL_VOLUME = 16383;
  static const uint32 MAX_PAN = 8192;
  static const uint32 CENTER_PAN = 8192;
  static const uint32 MAX_EXPRESSION = 16383;
  static const uchar MAX_REVERB_SEND_LEVEL = 127;

  //////////////////////////////////////RPN,NRPN
  static const int PARAMETER_NULL = 0;
  static const int PARAMETER_RPN  = 1;
  static const int PARAMETER_NRPN = 2;
  
private:
  bool __isMuted;

  static InstrumentName* instrumentName;
  static void initInstrumentName ();
  static pthread_once_t isInstrumentInited;
  const char* instrument;//$B3Z4oL>!#(B

  uchar useForRhythmPart;
  MIDI::PORT port;//CHANNEL MSG Rx PORT
  //MODEL ID=0x42 0x00 0x01 0x??(0x00-0x0F,0x10-0x1F) 0x0?(0,1=A,B port)
  
  bool lastNote[MIDI::MAX_NOTES];
  bool note[MIDI::MAX_NOTES];//$B$I$N(Bkey$B$,(Bnote on or off$B$+(B?
  double noteOnMachineTicks[MIDI::MAX_NOTES];//note on$B$7$?;~$N(Bmachine ticks
  uchar velocity[MIDI::MAX_NOTES];//note on$B$N(Bvelocity
  uchar maxVelocity;
  //note on$B$N(Bvelocity$B$N:GBg$NJ*$G0l2s$K(Bdump$B$9$k;~$N:GBg$r3P$($F$*$/!#(B
  //drum map$B$N;~$K(BLCD$B$rI=<($9$k;~$K;HMQ!#(B
  uchar lastMaxVelocity;
  //note on$B$N(Bvelocity$B$N:GBg$NJ*$G(Bdump,sync$B$7$?;~$NJ*$r3P$($F$*$/!#(B
  //drum map$B$N;~$K(BLCD$B$rI=<($9$k;~$K;HMQ!#(B
  uint32 lastTicks;//$B0lHV:G8e$KJQ2=$7$?;~4V!#(B(MIDI ticks)
  uint32 lastMachineTicks;//$B0lHV:G8e$KJQ2=$7$?;~4V!#(B(machine ticks)
  //$BC10L$O(Bmachine ticks$B$G$"$k$3$H$KCm0U!#(BSong$B$NJ}$N(Bticks$B$G$O$J$$!#(B
  //Song.rate$B$+$i7W;;=PMh$k!#(B
  int notesOnNumber;//note on$B$7$F$$$k?t!#(B
  
  uchar polyphonicKeyPressure[MIDI::MAX_NOTES];//key pressure 0xAn note dd
  uchar channelPressure;//channel pressure 0xDn dd
  uchar program;//program number 0xCn pp
  uchar bankMSB;//bank MSB 0xBn 0x00 dM
  uchar bankLSB;//bank LSB 0xBn 0x20 dL
  
  //pitch bend
  int32 pitchBend;//piitch bend change 0xEn LSB MSB

  ///0xBn
  uint32 modulationWheel;//modulation wheel  0xBn 0x01 0xdM 7bit
  //modulation wheel  0xBn 0x21 0xdL 7bit
  uint32 channelVolume;//channel volume 0xBn 0x07 0xdM
  //channel volume 0xBn 0x27 0xdL
  //static const uint32 CENTER_PAN=((0x40 << 7)+0x40);
  uint32 pan;//pan 0xBn 0x0A 0xdM 7bit
  //pan 0xBn 0x2A 0xdM 7bit
  uint32 expression;//expression controller 0xBn 0x0B 0xdM 7bit
  //expression controller 0xBn 0x2B 0xdM 7bit
  uchar damperPedal;//damper pedal(hold 1) 0xBn 0x40 dd 7bit
  uchar portamentSwitch;//portamento on/off 0xBn 0x41 dd
  uchar sostenute;//sostenute 0xBn 0x42 dd 7bit
  uchar softPedal;//soft pedal 0xBn 0x43 dd 7bit
  uchar reverbSendLevel;//effect1 depth(reverb send level) 0xBn 0x5B dd
  uchar chorusSendLevel;//effect3 depth(chorus send level) 0xBn 0x5D dd
  uchar delaySendLevel;//effect4 depth (celeste (detune)depth) 0xBn 0x5E dd

  //////////////////////////////////////RPN,NRPN
  int parameterKind;//$B8=:_(Brpn$B$+(Bnrpn$B$+(Bnull$B$+(B?
  uchar rpnMSB;
  uchar rpnLSB;
  uchar nrpnMSB;
  uchar nrpnLSB;

  ///////////////SysEx
  char vibrateRate;
  char vibrateDepth;
  char vibrateDelay;
  char tvfCutOff;
  char tvfResonance;
  char envelopeAttackTime;
  char envelopeDecayTime;
  char envelopeReleaseTime;
  
  uchar toneNumber;//CC#00 value 0-12 0x40 0x1? 0x00
  uchar RxChannel;//$B<u?.%A%c%s%M%k(B 0x40 0x1? 0x02//!!!!!!!!!!!!!!!!!!!!
  uchar RxPitchBend;//0x40 0x1? 0x03
  uchar RxChPressure;//0x40 0x1? 0x04
  uchar RxProgramChange;//0x40 0x1? 0x05
  uchar RxCongrolChange;//0x40 0x1? 0x06
  uchar RxPolyPressure;//0x40 0x1? 0x07
  uchar RxNoteMessage;//0x40 0x1? 0x08
  uchar RxRPN;//0x40 0x1? 0x09//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  uchar RxNRPN;//0x40 0x1? 0x0A//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  uchar RxModulation;//0x40 0x1? 0x0B
  uchar RxVolume;//0x40 0x1? 0x0C
  uchar RxPanpot;//0x40 0x1? 0x0D
  uchar RxExpression;//0x40 0x1? 0x0E
  uchar RxHold1;//0x40 0x1? 0x0F
  uchar RxPortamento;//0x40 0x1? 0x10
  uchar RxSostenute;//0x40 0x1? 0x11
  uchar RxSoft;//0x40 0x1? 0x12
  uchar MonoPolyMode;//0x40 0x1? 0x13
  uchar assignMode;//0x40 0x1? 0x14

  uchar RxBankSelect;//0x40 0x1? 0x23//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  uchar RxBandSelectLSB;//0x40 0x1? 0x24//!!!!!!!!!!!!!!!!!!!!!!!!!!!

public:
  Part ();//constructor
  ~Part ();

  //////////////////////////////////////////////////////////////////////
  void reset (uchar defaltPort, uchar receivedChannel);//$BJQ?t$N(Breset
  void reset ();//$BJQ?t$N(Breset
  uchar getNoteOnVelocity (const SongStatus& nowSongState) const;
  uchar getMaxVelocity (const SongStatus& nowSongStatus) const;
  //$BJL$N7W;;J}K!$G$d$C$F$_$k!#(B
  uchar getNoteOnVelocity2 (const SongStatus& nowSongState) const;
  uchar getMaxVelocity2 (const SongStatus& nowSongStatus) const;
  
  bool isUseForRhythmPart (SequencerStatus::DeviceMode deviceMode) const
    {
      if ((deviceMode & SequencerStatus::DEVICE_XG_ON) != 0)
	{//XG(Yamaha)
	  return ((bankMSB == 126) ||(bankMSB == 127));
	}
      else
	{//GM or 88 or 88Pro(Rolland)
	  return (useForRhythmPart == MAP_OFF);
	}
      return false;
    };
  
  //(port__, channel__)$B$r<u?.$7$F$$$k$+(B?
  bool isRecieve (MIDI::PORT port__, uchar channel__) const
    {
      return ((port == port__) && (RxChannel == channel__));
    };

  void setMute (bool isMuted__) {__isMuted = isMuted__;};
  void setPort (MIDI::PORT port__)
    {
      assert (port__ >= 0);
      assert (port__ < MIDI::MAX_PORT);
      port = port__;
    };
  
  void setMaxVelocity (uchar maxVelocity__)
    {maxVelocity = maxVelocity__;}
  void setLastMaxVelocity (uchar lastMaxVelocity__)
    {lastMaxVelocity = lastMaxVelocity__;}

  void setUseForRhythmPart (uchar useForRhythmPart__)
    {useForRhythmPart = useForRhythmPart__;};
  void setRxChannel (uchar RxChannel__) {RxChannel = RxChannel__;};
  
  
  ////////////////////////////////////////////////////////////access method
  bool isMuted () const {return __isMuted;};
  const char* getInstrumentName () const {return instrument;};
  uchar getUseForRhythmPart () const {return useForRhythmPart;};
  MIDI::PORT getPort () const {return port;};
  const bool getLastNote (uchar noteNumber) const
    {return lastNote[noteNumber];};
  const bool getNote (uchar noteNumber) const {return note[noteNumber];};
  double getNoteOnMachineTicks (uchar noteNumber) const
    {return noteOnMachineTicks[noteNumber];};
  uchar getMaxVelocity () const {return maxVelocity;};
  uchar getLastMaxVelocity () const {return lastMaxVelocity;};
  uint32 getLastTicks () const {return lastTicks;};
  uint32 getLastMachineTicks () const {return lastMachineTicks;};
  int getNotesOnNumber () const {return notesOnNumber;};
  uchar getPolyphonicKeyPressure (uchar noteNumber) const
    {return polyphonicKeyPressure[noteNumber];};
  uchar getChannelPressure () const {return channelPressure;};
  uchar getProgram () const {return program;};
  uchar getBankMSB () const {return bankMSB;};
  uchar getBankLSB () const {return bankLSB;};
  int32 getPitchBend () const {return pitchBend;};

  //Bx
  uint32 getModulationWheel () const {return modulationWheel;};
  uint32 getChannelVolume () const {return channelVolume;};
  uint32 getPan () const {return pan;};
  uint32 getExpression () const {return expression;};
  uchar getDamperPedal () const {return damperPedal;};
  uchar getPortamentSwitch () const {return portamentSwitch;};
  uchar getSostenute () const {return sostenute;}; 
  uchar getSoftPedal () const {return softPedal;};
  uchar getReverbSendLevel () const {return reverbSendLevel;}; 
  uchar getChorusSendLevel () const {return chorusSendLevel;};
  uchar getDelaySendLevel () const {return delaySendLevel;};

  //////////////////////////////////////RPN,NRPN
  int getParameterKind () const {return parameterKind;};
  uchar getRpnMSB () const {return rpnMSB;};
  uchar getRpnLSB () const {return rpnLSB;};
  uchar getNrpnMSB () const {return nrpnMSB;};
  uchar getNrpnLSB () const {return nrpnLSB;};

  //SysEx
  char getVibrateRate () const {return vibrateRate;};
  char getVibrateDepth () const {return vibrateDepth;};
  char getVibrateDelay () const {return vibrateDelay;};
  char getTvfCutOff () const {return tvfCutOff;};
  char getTvfResonance () const {return tvfResonance;};
  char getEnvelopeAttackTime () const {return envelopeAttackTime;};
  char getEnvelopeDecayTime () const {return envelopeDecayTime;};
  char getEnvelopeReleaseTime () const {return envelopeReleaseTime;};

  uchar getToneNumber () const {return toneNumber;};
  uchar getRxChannel () const {return RxChannel;};
  uchar getRxPitchBend () const {return RxPitchBend;};
  uchar getRxChPressure () const {return RxChPressure;};
  uchar getRxProgramChange () const {return RxProgramChange;};
  uchar getRxCongrolChange () const {return RxCongrolChange;};
  uchar getRxPolyPressure () const {return RxPolyPressure;};
  uchar getRxNoteMessage () const {return RxNoteMessage;};
  uchar getRxRPN () const {return RxRPN;};
  uchar getRxNRPN () const {return RxNRPN;};
  uchar getRxModulation () const {return RxModulation;};
  uchar getRxVolume () const {return RxVolume;};
  uchar getRxPanpot () const {return RxPanpot;};
  uchar getRxExpression () const {return RxExpression;};
  uchar getRxHold1 () const {return RxHold1;};
  uchar getRxPortamento () const {return RxPortamento;};
  uchar getRxSostenute () const {return RxSostenute;};
  uchar getRxSoft () const {return RxSoft;};
  uchar getMonoPolyMode () const {return MonoPolyMode;};
  uchar getAssignMode () const {return assignMode;};

  uchar getRxBankSelect () const {return RxBankSelect;};
  uchar getRxBandSelectLSB () const {return RxBandSelectLSB;};
  
public:
  ///////////////////////////////////////////////////////////trace midi status
  void traceNoteOff (uchar note__);
  void traceNoteOn (uchar note__, uchar velocity__, double machineTicks);
  void tracePolyphonicKeyPressure (uchar note__, uchar keyPressure);
  void traceControlChange (uchar control, uchar value);
private:
    //aux methods for traceControlChange
    void traceDataEntry (bool isMSB, uchar value);
    void traceAllSoundOff ();
    void traceResetAllController ();
public:
  void traceProgramChange (uchar program__,
			   SequencerStatus::DeviceMode deviceMode);
  void traceChannelPressure (uchar channelPressure__);
  void tracePitchBendChange (uchar LSB, uchar MSB);
};
#endif //Part_h
