/*
    SongFileItem.cc
  
    Copyright (C) 1998, 1999 Isaku Yamahata

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "SongFileItem.h"
#include "IO/MainIONull.h"
#include "Song/SongFactory.h"
#include "ServerCommand.h"
#include "SongManager.h"

void SongFileItem::setPlayTime (uint32 playTime__)
{
  assert (!isPlayTimeSet ());
  __isPlayTimeSet = true;
  playTime = playTime__;
}

void SongFileItem::setTitle (const string& title__)
{
  assert (!isTitleSet ());
  __isTitleSet = true;
  title = title__;
}

void SongFileItem::set (const Song* song)
{
  setPlayTime (song->getPlayTime ());
  setTitle (song->getTitle ());
}

bool SongFileItem::search ()
{
  assert (!isPlayTimeSet ());
  assert (!isTitleSet ());
  
  State dummyState;
  Option dummyOption;
  dummyOption.isQuickStart = true;
#if 0
  try
    {
#endif
      ServerCommand* serverCommand = new ServerCommand;
      SongManager* songManager = new SongManager ();
      WidiController dummyWidiController (serverCommand, songManager);
      MainIO* dummyMainIONull = new MainIONull (&dummyState,
						&dummyWidiController,
						&dummyOption);
#if 0
    }
    catch ()
    {
    }
#endif
  SongCommand dummySongCommand;

  bool isSearchSuccessed = false;
  Song* song = SongFactory::newSong (fileName, &dummyState, &dummySongCommand,
				     *dummyMainIONull, &dummyOption);
  if (song != NULL)
    {
      isSearchSuccessed = true;
      set (song);
    }
  
  delete song;
  delete dummyMainIONull;
  delete serverCommand;
  delete songManager;
  
  return isSearchSuccessed;
}



